package ctbrec.ui.tabs.recorded;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.time.Instant;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.ReentrantLock;
import java.util.stream.Collectors;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import com.google.common.eventbus.Subscribe;

import ctbrec.Config;
import ctbrec.GlobalThreadPool;
import ctbrec.Model;
import ctbrec.StringUtil;
import ctbrec.event.EventBusHolder;
import ctbrec.recorder.Recorder;
import ctbrec.sites.Site;
import ctbrec.ui.AutosizeAlert;
import ctbrec.ui.JavaFxModel;
import ctbrec.ui.PreviewPopupHandler;
import ctbrec.ui.action.AbstractPortraitAction.PortraitChangedEvent;
import ctbrec.ui.action.MarkForLaterRecordingAction;
import ctbrec.ui.action.PlayAction;
import ctbrec.ui.action.SetPortraitAction;
import ctbrec.ui.action.StartRecordingAction;
import ctbrec.ui.controls.CustomMouseBehaviorContextMenu;
import ctbrec.ui.controls.DateTimeCellFactory;
import ctbrec.ui.controls.Dialogs;
import ctbrec.ui.controls.SearchBox;
import ctbrec.ui.controls.autocomplete.AutoFillTextField;
import ctbrec.ui.controls.autocomplete.ObservableListSuggester;
import ctbrec.ui.menu.ModelMenuContributor;
import ctbrec.ui.tabs.TabSelectionListener;
import javafx.beans.property.SimpleObjectProperty;
import javafx.beans.property.SimpleStringProperty;
import javafx.beans.property.StringPropertyBase;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.event.ActionEvent;
import javafx.geometry.Insets;
import javafx.geometry.Point2D;
import javafx.geometry.Pos;
import javafx.scene.Cursor;
import javafx.scene.control.Alert;
import javafx.scene.control.Button;
import javafx.scene.control.CheckMenuItem;
import javafx.scene.control.ContextMenu;
import javafx.scene.control.Label;
import javafx.scene.control.ScrollPane;
import javafx.scene.control.SelectionMode;
import javafx.scene.control.Tab;
import javafx.scene.control.TableCell;
import javafx.scene.control.TableColumn;
import javafx.scene.control.TableColumn.SortType;
import javafx.scene.control.TableRow;
import javafx.scene.control.TableView;
import javafx.scene.control.TextField;
import javafx.scene.control.Tooltip;
import javafx.scene.control.cell.PropertyValueFactory;
import javafx.scene.image.Image;
import javafx.scene.input.ContextMenuEvent;
import javafx.scene.input.KeyCode;
import javafx.scene.input.KeyEvent;
import javafx.scene.input.MouseButton;
import javafx.scene.input.MouseEvent;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.FlowPane;
import javafx.scene.layout.HBox;
import javafx.scene.layout.Priority;
import javafx.util.Callback;

public abstract class AbstractRecordedModelsTab extends Tab implements TabSelectionListener {
    private static final Logger LOG = LoggerFactory.getLogger(AbstractRecordedModelsTab.class);
    private static final Image SILHOUETTE = new Image(AbstractRecordedModelsTab.class.getResourceAsStream("/silhouette_256.png"));
    protected static final String STYLE_ALIGN_CENTER = "-fx-alignment: CENTER;";

    protected ReentrantLock lock = new ReentrantLock();
    protected ObservableList<JavaFxModel> observableModels = FXCollections.observableArrayList();
    protected ObservableList<JavaFxModel> filteredModels = FXCollections.observableArrayList();

    protected TableView<JavaFxModel> table = new TableView<>();
    protected List<TableColumn<JavaFxModel, ?>> columns = new ArrayList<>();
    protected LoadingCache<Model, Image> portraitCache = CacheBuilder.newBuilder()
            .expireAfterAccess(1, TimeUnit.DAYS)
            .maximumSize(1000)
            .build(CacheLoader.from(AbstractRecordedModelsTab::loadModelPortrait));

    protected AutoFillTextField modelInputField;
    protected List<Site> sites;
    protected Recorder recorder;

    protected HBox addModelBox = new HBox(5);
    protected HBox filterContainer = new HBox(5);
    protected Label modelLabel = new Label("Model");
    protected Button addModelButton = new Button("Record");
    protected Button checkModelAccountExistance = new Button("Check URLs");
    protected TextField filter;

    protected FlowPane grid = new FlowPane();
    protected ScrollPane scrollPane = new ScrollPane();
    protected ContextMenu popup;

    protected Config config;

    AbstractRecordedModelsTab(String text) {
        super(text);
        config = Config.getInstance();
        registerPortraitListener();
    }

    protected void registerPortraitListener() {
        EventBusHolder.BUS.register(this);
    }

    @Subscribe
    public void portraitChanged(PortraitChangedEvent e) {
        portraitCache.invalidate(e.getModel());
        if (table != null) {
            table.refresh();
        }
    }

    protected void createGui() {
        grid.setPadding(new Insets(5));
        grid.setHgap(5);
        grid.setVgap(5);

        scrollPane.setContent(grid);
        scrollPane.setFitToHeight(true);
        scrollPane.setFitToWidth(true);
        BorderPane.setMargin(scrollPane, new Insets(5));

        table.setEditable(true);
        if (!config.getSettings().showGridLinesInTables) {
            table.setStyle("-fx-table-cell-border-color: transparent;");
        }
        table.getSelectionModel().setSelectionMode(SelectionMode.MULTIPLE);
        var previewPopupHandler = new PreviewPopupHandler(table);
        table.setRowFactory(tableview -> {
            TableRow<JavaFxModel> row = new TableRow<>();
            row.addEventHandler(MouseEvent.ANY, previewPopupHandler);
            return row;
        });
        table.setItems(observableModels);
        table.addEventHandler(ContextMenuEvent.CONTEXT_MENU_REQUESTED, event -> {
            popup = createContextMenu();
            if (popup != null) {
                popup.show(table, event.getScreenX(), event.getScreenY());
            }
            event.consume();
        });
        table.addEventHandler(MouseEvent.MOUSE_PRESSED, event -> {
            if (popup != null) {
                popup.hide();
            }
        });
        table.addEventHandler(KeyEvent.KEY_PRESSED, event -> {
            List<JavaFxModel> selectedModels = table.getSelectionModel().getSelectedItems();
            if (event.getCode() == KeyCode.DELETE) {
                stopAction(selectedModels);
            } else {
                jumpToNextModel(event.getCode());
            }
        });

        scrollPane.setContent(table);

        checkModelAccountExistance.setPadding(new Insets(5));
        checkModelAccountExistance.setTooltip(new Tooltip("Go over all model URLs and check, if the account still exists"));
        HBox.setMargin(checkModelAccountExistance, new Insets(0, 0, 0, 20));

        modelLabel.setPadding(new Insets(5, 0, 0, 0));
        ObservableList<String> suggestions = FXCollections.observableArrayList();
        sites.forEach(site -> suggestions.add(site.getClass().getSimpleName()));
        modelInputField = new AutoFillTextField(new ObservableListSuggester(suggestions));
        modelInputField.minWidth(150);
        modelInputField.prefWidth(600);
        HBox.setHgrow(modelInputField, Priority.ALWAYS);
        modelInputField.setPromptText("e.g. MyFreeCams:ModelName or an URL like https://chaturbate.com/modelname/");
        modelInputField.onActionHandler(this::addModel);
        modelInputField.setTooltip(new Tooltip("To add a model enter SiteName:ModelName\n" + "press ENTER to confirm a suggested site name"));
        BorderPane.setMargin(addModelBox, new Insets(5));
        addModelButton.setOnAction(this::addModel);
        addModelButton.setPadding(new Insets(5));
        addModelBox.getChildren().addAll(modelLabel, modelInputField, addModelButton, checkModelAccountExistance);

        filterContainer.setPadding(new Insets(0));
        filterContainer.setAlignment(Pos.CENTER_RIGHT);
        filterContainer.minWidth(100);
        filterContainer.prefWidth(150);
        HBox.setHgrow(filterContainer, Priority.ALWAYS);
        filter = new SearchBox(false);
        filter.minWidth(100);
        filter.prefWidth(150);
        filter.setPromptText("Filter");
        filter.textProperty().addListener((observableValue, oldValue, newValue) -> {
            String q = filter.getText();
            lock.lock();
            try {
                filter(q);
            } finally {
                lock.unlock();
            }
        });
        filter.getStyleClass().remove("search-box-icon");

        var columnSelection = new Button("⚙");
        columnSelection.setOnAction(this::showColumnSelection);
        columnSelection.setTooltip(new Tooltip("Select columns"));
        columnSelection.prefHeightProperty().bind(filter.prefHeightProperty());
        columnSelection.prefWidthProperty().bind(columnSelection.prefHeightProperty());

        filterContainer.getChildren().addAll(columnSelection, filter);
        addModelBox.getChildren().add(filterContainer);
    }

    protected void addPreviewColumn(int columnIdx) {
        TableColumn<JavaFxModel, String> preview = addTableColumn("preview", "🎥", columnIdx, 35);
        preview.setCellValueFactory(cdf -> new SimpleStringProperty("  ▶  "));
        preview.setEditable(false);
        if (!Config.getInstance().getSettings().livePreviews) {
            preview.setVisible(false);
        }
    }

    protected void addPortraitColumn(int columnIdx) {
        TableColumn<JavaFxModel, Image> portrait = addTableColumn("portrait", "Portrait", columnIdx, 80);
        portrait.setCellValueFactory(param -> {
            Model mdl = param.getValue().getDelegate();
            SimpleObjectProperty<Image> imgProperty = new SimpleObjectProperty<>();
            Image image = portraitCache.getIfPresent(mdl);
            if (image == null) {
                GlobalThreadPool.submit(() -> portraitCache.put(mdl, loadModelPortrait(mdl)));
            }
            imgProperty.set(image);
            return imgProperty;
        });
        portrait.setCellFactory(param -> new ImageTableCell());
        portrait.setEditable(false);
    }

    protected void addModelColumn(int columnIdx) {
        TableColumn<JavaFxModel, ModelName> name = addTableColumn("name", "Model", columnIdx, 200);
        name.setCellValueFactory(param -> {
            var modelName = new ModelName(param.getValue(), recorder);
            return new SimpleObjectProperty<>(modelName);
        });
        name.setCellFactory(param -> new ModelNameTableCell(recorder));
        name.setEditable(false);
    }

    protected void addUrlColumn(int columnIdx) {
        TableColumn<JavaFxModel, String> url = addTableColumn("url", "URL", columnIdx, 400);
        url.setCellValueFactory(new PropertyValueFactory<>("url"));
        url.setCellFactory(new ClickableCellFactory<>());
        url.setEditable(false);
    }

    protected void addAddedTimestampColumn(int columnIdx) {
        TableColumn<JavaFxModel, Instant> tc = addTableColumn("addedTimestamp", "added at", columnIdx, 400);
        tc.setCellFactory(new DateTimeCellFactory<>());
        tc.setCellValueFactory(param -> new SimpleObjectProperty<>(param.getValue().getAddedTimestamp()));
        tc.setPrefWidth(150);
        tc.setEditable(false);
        tc.setStyle(STYLE_ALIGN_CENTER);
    }

    protected void addNotesColumn(int columnIdx) {
        TableColumn<JavaFxModel, String> notes = addTableColumn("notes", "Notes", columnIdx, 400);
        notes.setCellValueFactory(cdf -> {
            JavaFxModel m = cdf.getValue();
            return new StringPropertyBase() {
                @Override
                public String getName() {
                    return "Model Notes";
                }

                @Override
                public Object getBean() {
                    return null;
                }

                @Override
                public String get() {
                    String modelNotes = Config.getInstance().getModelNotes(m);
                    return modelNotes;
                }
            };
        });
        notes.setEditable(false);
    }

    abstract void stopAction(List<JavaFxModel> selectedModels);

    protected <T> TableColumn<JavaFxModel, T> addTableColumn(String id, String text, int index, int width) {
        TableColumn<JavaFxModel, T> tc = new TableColumn<>(text);
        tc.setId(id);
        tc.setText(text);
        tc.setUserData(index);
        tc.setPrefWidth(width);
        tc.setStyle("-fx-alignment: CENTER-LEFT;");
        columns.add(tc);
        addTableColumnIfEnabled(tc);
        return tc;
    }

    protected void addTableColumnIfEnabled(TableColumn<JavaFxModel, ?> tc) {
        if(isColumnEnabled(tc)) {
            table.getColumns().add(tc);
        }
    }

    protected ContextMenu createContextMenu() {
        List<Model> selectedModels = table.getSelectionModel().getSelectedItems().stream().map(JavaFxModel::getDelegate).collect(Collectors.toList());
        if (selectedModels.isEmpty()) {
            return null;
        }

        ContextMenu menu = new CustomMouseBehaviorContextMenu();

        ModelMenuContributor.newContributor(getTabPane(), Config.getInstance(), recorder) //
        .withStartStopCallback(m -> getTabPane().setCursor(Cursor.DEFAULT)) //
        .removeModelAfterIgnore(true) //
        .withPortraitCallback(m -> {
            portraitCache.invalidate(m);
            table.refresh();
        })
        .afterwards(table::refresh) //
        .contributeToMenu(selectedModels, menu);

        return menu;
    }

    protected void addModel(ActionEvent e) {
        String input = modelInputField.getText().trim();
        if (StringUtil.isBlank(input)) {
            return;
        }

        if (input.startsWith("http")) {
            addModelByUrl(input);
        } else {
            addModelByName(input);
        }
    }

    protected void addModelByUrl(String url) {
        for (Site site : sites) {
            var newModel = site.createModelFromUrl(url);
            if (newModel != null) {
                if (getMarkModelsForLaterRecording()) {
                    new MarkForLaterRecordingAction(getTabPane(), List.of(newModel), true, recorder).execute();
                } else {
                    new StartRecordingAction(getTabPane(), List.of(newModel), recorder).execute();
                }
                return;
            }
        }

        Dialogs.showError(getTabPane().getScene(), "Unknown URL format",
                "The URL you entered has an unknown format or the function does not support this site, yet", null);
    }

    abstract boolean getMarkModelsForLaterRecording();

    protected void addModelByName(String siteModelCombo) {
        String[] parts = siteModelCombo.trim().split(":");
        if (parts.length != 2) {
            Dialogs.showError(getTabPane().getScene(), "Wrong input format", "Use something like \"MyFreeCams:ModelName\"", null);
            return;
        }

        String siteName = parts[0];
        String modelName = parts[1];
        for (Site site : sites) {
            if (Objects.equals(siteName.toLowerCase(), site.getClass().getSimpleName().toLowerCase())) {
                var newModel = site.createModel(modelName);
                if (getMarkModelsForLaterRecording()) {
                    new MarkForLaterRecordingAction(getTabPane(), List.of(newModel), true, recorder).execute();
                } else {
                    new StartRecordingAction(getTabPane(), List.of(newModel), recorder).execute();
                }
                return;
            }
        }

        Alert alert = new AutosizeAlert(Alert.AlertType.ERROR, getTabPane().getScene());
        alert.setTitle("Unknown site");
        alert.setHeaderText("Couldn't add model");
        alert.setContentText("The site you entered is unknown");
        alert.showAndWait();
    }

    protected void jumpToNextModel(KeyCode code) {
        if (!table.getItems().isEmpty() && (code.isLetterKey() || code.isDigitKey())) {
            // determine where to start looking for the next model
            var startAt = getJumpToStartIndex();

            String c = code.getChar().toLowerCase();
            int i = startAt;
            do {
                JavaFxModel current = table.getItems().get(i);
                if (current.getName().toLowerCase().replaceAll("[^0-9a-z]", "").startsWith(c)) {
                    table.getSelectionModel().clearAndSelect(i);
                    table.scrollTo(i);
                    break;
                }

                i++;
                if (i >= table.getItems().size()) {
                    i = 0;
                }
            } while (i != startAt);
        }
    }

    protected int getJumpToStartIndex() {
        var startAt = 0;
        if (table.getSelectionModel().getSelectedIndex() >= 0) {
            startAt = table.getSelectionModel().getSelectedIndex() + 1;
            if (startAt >= table.getItems().size()) {
                startAt = 0;
            }
        }
        return startAt;
    }

    protected void filter(String filter) {
        lock.lock();
        try {
            if (StringUtil.isBlank(filter)) {
                observableModels.addAll(filteredModels);
                filteredModels.clear();
                return;
            }

            String[] tokens = filter.split(" ");
            observableModels.addAll(filteredModels);
            filteredModels.clear();
            for (var i = 0; i < table.getItems().size(); i++) {
                var sb = new StringBuilder();
                for (TableColumn<JavaFxModel, ?> tc : table.getColumns()) {
                    Object cellData = tc.getCellData(i);
                    if (cellData != null) {
                        var content = cellData.toString();
                        sb.append(content).append(' ');
                    }
                }
                var searchText = sb.toString();

                var tokensMissing = false;
                for (String token : tokens) {
                    if (!searchText.toLowerCase().contains(token.toLowerCase())) {
                        tokensMissing = true;
                        break;
                    }
                }
                if (tokensMissing) {
                    JavaFxModel filteredModel = table.getItems().get(i);
                    filteredModels.add(filteredModel);
                }
            }
            observableModels.removeAll(filteredModels);
        } finally {
            lock.unlock();
        }
    }

    abstract String getSortColumn();
    abstract void setSortColumn(String column);
    abstract String getSortType();
    abstract void setSortType(String sortType);
    abstract String[] getColumnIds();
    abstract void setColumnIds(String[] ids);
    abstract double[] getColumnWidths();
    abstract void setColumnWidths(double[] widths);
    abstract List<String> getDisabledColumns();

    public void saveState() {
        if (!table.getSortOrder().isEmpty()) {
            TableColumn<JavaFxModel, ?> col = table.getSortOrder().get(0);
            setSortColumn(col.getText());
            setSortType(col.getSortType().toString());
        }
        int tableColumns = table.getColumns().size();
        var columnWidths = new double[tableColumns];
        var columnIds = new String[tableColumns];
        for (var i = 0; i < columnWidths.length; i++) {
            columnWidths[i] = table.getColumns().get(i).getWidth();
            columnIds[i] = table.getColumns().get(i).getId();
        }
        setColumnWidths(columnWidths);
        setColumnIds(columnIds);
    }

    protected void restoreState() {
        restoreColumnOrder();
        restoreColumnWidths();
        restoreSorting();
    }


    private void restoreSorting() {
        String sortCol = getSortColumn();
        if (StringUtil.isNotBlank(sortCol)) {
            for (TableColumn<JavaFxModel, ?> col : table.getColumns()) {
                if (Objects.equals(sortCol, col.getText())) {
                    col.setSortType(SortType.valueOf(getSortType()));
                    table.getSortOrder().clear();
                    table.getSortOrder().add(col);
                    break;
                }
            }
        }
    }

    private void restoreColumnOrder() {
        String[] columnIds = getColumnIds();
        ObservableList<TableColumn<JavaFxModel,?>> tableColumns = table.getColumns();
        for (var i = 0; i < columnIds.length; i++) {
            for (var j = 0; j < table.getColumns().size(); j++) {
                if(Objects.equals(columnIds[i], tableColumns.get(j).getId())) {
                    TableColumn<JavaFxModel, ?> col = tableColumns.get(j);
                    tableColumns.remove(j); // NOSONAR
                    tableColumns.add(Math.min(i, tableColumns.size()), col);
                }
            }
        }
    }

    private void restoreColumnWidths() {
        double[] columnWidths = getColumnWidths();
        if (columnWidths != null && columnWidths.length == table.getColumns().size()) {
            for (var i = 0; i < columnWidths.length; i++) {
                table.getColumns().get(i).setPrefWidth(columnWidths[i]);
            }
        }
    }

    protected class ClickableCellFactory<S, T> implements Callback<TableColumn<S, T>, TableCell<S, T>> {
        @Override
        public TableCell<S, T> call(TableColumn<S, T> param) {
            TableCell<S, T> cell = new TableCell<>() {
                @Override
                protected void updateItem(Object item, boolean empty) {
                    setText(empty ? "" : Objects.toString(item));
                }
            };

            cell.addEventFilter(MouseEvent.MOUSE_CLICKED, event -> {
                if (event.getButton() == MouseButton.PRIMARY && event.getClickCount() == 2) {
                    JavaFxModel selectedModel = table.getSelectionModel().getSelectedItem();
                    if (selectedModel != null) {
                        new PlayAction(table, selectedModel).execute();
                    }
                }
            });
            return cell;
        }
    }

    protected static Image loadModelPortrait(Model model) {
        String portraitId = Config.getInstance().getSettings().modelPortraits.get(model.getUrl());
        if (StringUtil.isNotBlank(portraitId)) {
            File configDir = Config.getInstance().getConfigDir();
            File portraitDir = new File(configDir, "portraits");
            File portraitFile = new File(portraitDir, portraitId + '.' + SetPortraitAction.FORMAT);
            try {
                return new Image(new FileInputStream(portraitFile));
            } catch (FileNotFoundException e) {
                LOG.error("Couldn't load portrait file {}", portraitFile, e);
            }
        }
        return SILHOUETTE;
    }

    protected void showColumnSelection(ActionEvent evt) {
        ContextMenu menu = new CustomMouseBehaviorContextMenu();
        for (TableColumn<JavaFxModel, ?> tc : columns) {
            var item = new CheckMenuItem(tc.getText());
            item.setSelected(isColumnEnabled(tc));
            menu.getItems().add(item);
            item.setOnAction(e -> {
                try {
                    if (item.isSelected()) {
                        getDisabledColumns().remove(tc.getText());
                        boolean added = false;
                        for (int i = table.getColumns().size() - 1; i >= 0; i--) {
                            TableColumn<JavaFxModel, ?> other = table.getColumns().get(i);
                            if (!other.isVisible()) {
                                continue;
                            }
                            int idx = (int) tc.getUserData();
                            int otherIdx = (int) other.getUserData();
                            if (otherIdx < idx) {
                                table.getColumns().add(i + 1, tc);
                                added = true;
                                break;
                            }
                        }
                        if (!added) {
                            table.getColumns().add(0, tc);
                        }
                    } else {
                        getDisabledColumns().add(tc.getText());
                        table.getColumns().remove(tc);
                    }
                } catch (Exception ex) {
                    LOG.error("Couldn't activate column {}", tc, ex);
                }
            });
        }
        Button src = (Button) evt.getSource();
        Point2D location = src.localToScreen(src.getTranslateX(), src.getTranslateY());
        menu.show(getTabPane().getScene().getWindow(), location.getX(), location.getY() + src.getHeight() + 5);
    }

    private boolean isColumnEnabled(TableColumn<JavaFxModel, ?> tc) {
        return !getDisabledColumns().contains(tc.getText());
    }


}
